/****************************************************************************/
/*! \page EnumsetLicense enumset License
*\verbatim
* Copyright (c) 2001 - 2003 Marcus Boerger.  All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions
* are met:
*
* 1. Redistributions of source code must retain the above copyright
*    notice, this list of conditions and the following disclaimer.
*
* 2. Redistributions in binary form must reproduce the above copyright
*    notice, this list of conditions and the following disclaimer in
*    the documentation and/or other materials provided with the
*    distribution.
*
* 3. All advertising materials mentioning features or use of this
*    software must display the following acknowledgment:
*    "This product includes software developed by Marcus Boerger
*    (http://marcus-boerger.de/)."
*
* 4. The name "Marcus Boerger" must not be used to endorse or promote
*    products derived from this software without prior written permission.
*
* 5. Redistributions of any form whatsoever must retain the following
*    acknowledgment:
*    "This product includes software developed by Marcus Boerger
*    (http://marcus-boerger.de/)."
*
* 6. This is only intended to be used with enum or scalar types.
*
* THIS SOFTWARE IS PROVIDED BY MARCUS BOERGER ``AS IS'' AND ANY
* EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
* PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL ``MARCUS BOERGER'' OR HIS
* CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
* EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
* PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
* PROFITS; OR BUSINESS INTERRUPTION)
* HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
* STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
* OF THE POSSIBILITY OF SUCH DAMAGE.
\endverbatim
*//* ====================================================================== */

/* ------------------------------------------------------------------------ */
/*!
 * \file      enumset.hpp
 * \ingroup   MBO
 * \author    Marcus Boerger (http://marcus-boerger.de/)
 *
 * \brief     template mbo::enumset: a set template that works with enum types.
 *
 * Name:      enumset.hpp
 *
 * The template class mbo::enumset has been tested under:
 * - Linux with GCC 2.96 and 3.2
 * - Cygwin GCC 3.2
 * - WinXP with MSVC.NET
 *
 * \note      This file is distributed under the following license 
 *            \ref EnumsetLicense.
 *
 * \note      You can download the actual version of this file here: \n
 *            http://marcus-boerger.de/cpp/enumset/enumset.tar.bz2
 *
 * \note      If you do not have doxygen you can also download this 
 *            documentation as a chm file here: \n
 *            http://marcus-boerger.de/cpp/enumset/doc/enumset.chm \n
 *            or view it online here: \n
 *            <a href="javascript:window.top.location='http://marcus-boerger.de/cpp/enumset/doc/html/'">marcus-boerger.de/cpp/enumset/doc/html/</a> \n
 *
 */
/* ------------------------------------------------------------------------ */

/*! \if enumset_main
 * \mainpage mbo::enumset
 * \copydoc enumset.hpp
 * \endif
 */
/* ------------------------------------------------------------------------ */

#ifndef _ENUMSET_HPP_
#define _ENUMSET_HPP_

#ifdef __MSC_VER
#pragma comment (exestr, "This software includes material by M.Boerger (http://marcus-boerger.de/)")
#endif

#include <stdarg.h>
#include <exception>

#ifndef _Ty_list_default
/*! default type internally used to represent the set of enum values
 * \see mbo::enumset
 */
#define _Ty_list_default size_t
#endif

#ifndef _Ty_args_default
/*! default type used for parameter transport in '...' function calls
 * \see mbo::enumset
 */
#define _Ty_args_default int
#endif

/* ------------------------------------------------------------------------ */
/*! \if enumset_main
 * contains template class mbo::enumset
 * \endif
 */
namespace mbo {

/*! \brief base for exceptions thrown in mbo::enumset
 */
class enumset_exception: public std::exception{};

/*! \brief exception thrown for illegal input to mbo::enumset 
 */
class input_range_exception: public enumset_exception {};

/*! \brief base for exceptions thrown in mbo::enumset::to_enum
 *
 * this exception is thrown when an mbo::enumset instance which does not hold 
 * exactly one value is going to be converted to its enum type
 */
class demap_value_exception: public enumset_exception {};

/*! \brief exception thrown in mbo::enumset conversion s
 *
 * this exception is thrown when an mbo::enumset instance which holds more 
 * than one set value is going to be converted to its enum type
 */
class multi_value_exception: public demap_value_exception {};

/*! \brief exception thrown in mbo::enumset conversion
 *
 * this exception is thrown when an empty mbo::enumset instance (one which 
 * holds none of the set values) is going to be converted to its enum type
 */
class empty_value_exception: public demap_value_exception {};

/*! \brief exception thrown on illegal mbo::enumset operators
 *
 * this is exception is thrown when an illegal operator
 * was accessed from within a class derived from mbo::enumset.
 */
class operator_na_exception: public enumset_exception {};

/* ------------------------------------------------------------------------ */
/*!
 * \brief    A set template that works with enum types.
 *
 * \note     This template class is distributed under the following license 
 *           \ref EnumsetLicense.
 *
 * operators: + - * = += -= *=
 *
 * - + : OR, union
 * - - : exception/minus
 * - * : AND, intersection
 * - ^ : XOR
 *
 * boolean operators:   < <= > >= == !=
 *
 * The boolean operators are overloaded to behave as known from set theory.
 * 
 * There are two possible number (scalar) conversions:
 * - you may want to convert the instance to the numberic value internally
 *   used by calling to_set.
 * - it is also possible to convert the instance back to the enum value by use
 *   of to_enum which requires the instance holds exactly one value.
 *
 * All operators accept same mbo::enumset type as well as any enum value of the 
 * enum type used to construct the class.
 *
 * \note   The maximum value in TEnum must be less than or equal to the number
 *         of bits in type _Ty_list which defaults to size_t (platform 
 *         dependant: 16, 32 or 64 bits).
 *
 * When using numeric constants with mbo::enumset derived by enum the compiler
 * generates warnings of type 'no match ... candidates'. The behaviour of
 * passing non enum types or non scalar types as template type is neither
 * intended nor tested.
 *
 * \note   The operator ~ is missing because the constructor cannot 
 *         be informed about all set members available.
 *
 * Example:
 \verbatim
  #include "enumset.hpp"

  typedef enum {a,b,c} TEnum;
  typedef mbo::enumset<TEnum> TSet;
 
  TSet es;       // []
 
  es += a;       // [a]    addition, union
  es  = es + b;  // [a,b]
  es += es;      // [a,b]
  es *= b;       // [b]    intersection
  es  = es + c;  // [b,c]
  es -= b;       // [c]    minus
  es.clear();    // []
  printf("%d", (int)es);   // since size_t converts to int automatically
 \endverbatim
 *
 * \note You can find more complex examples in \ref enumset_example_1
 *       and \ref enumset_example_2.
 *
 * \param _TEnum    Specifies the enum type used to construct the class.
 * \param _Ty_list  Defaults to \c size_t. You may want to either override
 *                  this type for selective template instances or globally
 *                  by defining the macro _Ty_list_default before including 
 *                  the header file enumset.hpp.\n
 *                  For example you may want to use the type \c long \c long
 *                  to be able to use 64 values if a) your compiler knows
 *                  that type and b) \c size_t only uses 4 bytes which allows
 *                  only 32 values.
 * \param _Ty_args  Defaults to \c int. It my be necessary to override the
 *                  type used to transport parameters in '...' functions.
 *                  Normally the compiler uses the type \c int for enum types
 *                  but most compilers allow to overwrite this behaviour.\n
 *                  It may be necessary to override this type for selective
 *                  classes or globally by defining the macro _Ty_args_default
 *                  before including the file enumset.hpp.
 */
template < typename _TEnum
         , typename _Ty_list  = _Ty_list_default
         , typename _Ty_args  = _Ty_args_default
         >
class enumset {
public:

	/*! the enum type used to instanciate the template
	 *
	 * see notes on template parameters
	 */
	typedef _TEnum   TEnum;
	
	/*! type internally used to represent the set of enum values
	 *
	 * see notes on template parameters
	 */
	typedef _Ty_list Ty_list;
	
	/*! type used for parameter transport in '...' function calls
	 *
	 * see notes on template parameters
	 */
	typedef _Ty_args Ty_args;
	
	/*! empty constructor
	 */
	inline enumset () : list(0)
	{
	}

	/*! copy constructor
	 */
	inline enumset (const enumset & set)
	{
		list = set.list;
	}

	/*! dynamic construction
	 *
	 * \param count   numer of enum values that follow
	 * \param ...     enum parameters, count times
	 */
	inline explicit enumset (size_t count, ...)
		throw(input_range_exception)
		: list(0)
	{
	    va_list   args;
	    Ty_list   val;

    	va_start(args, count);
    	while (count--) {
    		val = (Ty_list)va_arg(args, Ty_args);
    		list |= map(val);
    	}
    	va_end(args);
	}

	/*! constructor using a const type reference
	 * \param value Value to store in the instance
	 */
	inline enumset (const TEnum & value)
		throw(input_range_exception)
	{
		list = map(value);
	}

	/*! Check whether or not a specific value is contained in the mbo::enumset
	 */
	inline bool contains(const TEnum & value) const
	{
		return (*this >= value);
	}

	/*! Check whether or not the value of another mbo::enumset is contained in the 
	 * mbo::enumset
	 */
	inline bool contains(const enumset & set) const
	{
		return (*this >= set);
	}

	/*! Number of single values held by the mbo::enumset
	 */
	inline int count() const
	{
	    register int i = 0;
	    register Ty_list p = 1;

        while (p) {
            i += (list & p) ? 1 : 0;
            p *=2;
        }
        return i;
	}

	/*! Clear the mbo::enumset (make it empty)
	 */
	inline enumset & clear()
	{
		list = 0;
		return *this;
	}

	/*! Assignment operator
	 */
	inline enumset & operator = (const TEnum & value)
		throw(input_range_exception)
	{
		list = map(value);
		return *this;
	}

	/*! Assignment operator
	 */
	inline enumset & operator = (const enumset & set)
	{
		list = set.list;
		return *this;
	}

	/*! Union/Append assignment
	 */
	inline enumset & operator += (const TEnum & value)
		throw(input_range_exception)
	{
		list |= map(value);
		return *this;
	}

	/*! Union/Append assignment
	 */
	inline enumset & operator += (const enumset & set)
	{
		list |= set.list;
		return *this;
	}

	/*! Union/Addition 
	 */
	inline enumset operator + (const TEnum & value) const
	{
		register enumset ret = enumset(*this);
		ret += value;
		return ret;
	}

	/*! Union/Addition 
	 */
	inline enumset operator + (const enumset & set) const
	{
		register enumset ret = enumset(*this);
		ret += set;
		return ret;
	}

	/*! Minus assignment
	 */
	inline enumset & operator -= (const TEnum &value)
		throw(input_range_exception)
	{
		list &= ~map(value);
		return *this;
	}

	/*! Minus assignment
	 */
	inline enumset & operator -= (const enumset &set)
	{
		list &= ~set.list;
		return *this;
	}

	/*! Exception
	 */
	inline enumset operator - (const TEnum & value) const
	{
		enumset ret = enumset(*this);
		ret -= value;
		return ret;
	}

	/*! Exception
	 */
	inline enumset operator - (const enumset & set) const
	{
		enumset ret = enumset(*this);
		ret -= set;
		return ret;
	}

	/*! Intersection assignment
	 */
	inline enumset & operator *= (const TEnum &value)
		throw(input_range_exception)
	{
		list &= map(value);
		return *this;
	}

	/*! Intersection assignment
	 */
	inline enumset & operator *= (const enumset &set)
	{
		list &= set.list;
		return *this;
	}

	/*! Intersection
	 */
	inline enumset operator * (const TEnum & value) const
	{
		register enumset ret = enumset(*this);
		ret *= value;
		return ret;
	}

	/*! Intersection
	 */
	inline enumset operator * (const enumset & set) const
	{
		register enumset ret = enumset(*this);
		ret *= set;
		return ret;
	}

	/*! XOR assignment
	 */
	inline enumset & operator ^= (const TEnum &value)
		throw(input_range_exception)
	{
		list ^= map(value);
		return *this;
	}

	/*! XOR assignment
	 */
	inline enumset & operator ^= (const enumset &set)
	{
		list ^= set.list;
		return *this;
	}

	/*! XOR
	 */
	inline enumset operator ^ (const TEnum & value) const
	{
		register enumset ret = enumset(*this);
		ret ^= value;
		return ret;
	}

	/*! XOR
	 */
	inline enumset operator ^ (const enumset & set) const
	{
		register enumset ret = enumset(*this);
		ret ^= set;
		return ret;
	}

	/*! Check whether or not the mbo::enumset is empty
	 */
	inline bool operator & (const TEnum & value) const
		throw(input_range_exception)
	{
		return (list & map(value)) > 0;
	}

	/*! Check whether or not the mbo::enumset is empty
	 */
	inline bool operator & (const enumset & set) const
	{
		return (list & set.list) > 0;
	}

	/*! EQ: Equivalence operator
	 */
	inline bool operator == (const TEnum & value) const
		throw(input_range_exception)
	{
		return list == map(value);
	}

	/*! EQ: Equivalence operator
	 */
	inline bool operator == (const enumset & set) const
	{
		return list == set.list;
	}

	/*! NE: NON-Equivalence operator
	 */
	inline bool operator != (const TEnum & value) const
		throw(input_range_exception)
	{
		return list != map(value);
	}

	/*! NE: NON-Equivalence operator
	 */
	inline bool operator != (const enumset & set) const
	{
		return list != set.list;
	}

	/*! LT: Less than operator
	 */
	inline bool operator < (const TEnum & value) const
		throw(input_range_exception)
	{
		register Ty_list comp = map(value);
		return ((list & comp) == list) && (list < comp);
	}

	/*! LT: Less than operator
	 */
	inline bool operator < (const enumset & set) const 
	{
		return ((list & set.list) == list) && (list < set.list);
	}

	/*! LE: Less than or Equal operator
	 */
	inline bool operator <= (const TEnum & value) const 
		throw(input_range_exception)
	{
		register Ty_list comp = map(value);
		return ((list & comp) == list) && (list <= comp);
	}

	/*! LE: Less than or Equal operator
	 */
	inline bool operator <= (const enumset & set) const 
	{
		return ((list & set.list) == list) && (list <= set.list);
	}

	/*! GT: Greater than operator
	 */
	inline bool operator > (const TEnum & value) const 
		throw(input_range_exception)
	{
		register Ty_list comp = map(value);
		return ((list & comp) == comp) && (list > comp);
	}

	/*! GT: Greater than operator
	 */
	inline bool operator > (const enumset & set) const 
	{
		return ((list & set.list) == set.list) && (list > set.list);
	}

	/*! GE: Greater than or Equal operator
	 */
	inline bool operator >= (const TEnum & value) const
		throw(input_range_exception)
	{
		register Ty_list comp = map(value);
		return ((list & comp) == comp) && (list >= comp);
	}

	/*! GE: Greater than or Equal operator
	 */
	inline bool operator >= (const enumset & set) const
	{
		return ((list & set.list) == set.list) && (list >= set.list);
	}

	/*! Conversion to a number
	 *
	 * \return internal representation
	 */
	inline Ty_list to_set() const
	{
		return list;
	}

	/*! \copydoc to_set
	 */	
	inline operator Ty_list() const
	{
		return list;
	}

	/*! Conversion to enum value
	 *
	 * When working with enum's this method returns a single enum value
	 * of the type used to instanciate this template class.
	 *
	 * The function requires the instance to hold exactly one
	 *      enum value. You may test this with the function \c count or catch
	 *      the thrown exception as illustrated in the following example:
	 *
	 \verbatim
#include <typeinfo>
#include <stdio.h>
#include "enumset.hpp"

enum TEnum {a=1,b,c};

typedef mbo::enumset<TEnum> TSet;

int main(int argc, void ** argv)
{
	try {
		TSet es;
		
		es = a;
		
		if (es.count()==1) {
			printf("es = %d\n", (TSet::TEnum)es);
		} else {
			printf("es is empty\n");
		}

		es +=b;
		printf("es = %d\n", (TSet::TEnum)es); // will display an exception
	}
	catch (mbo::demap_value_exception &e) {
		printf("illegal TSet value, count != 1\n");
	}
	catch (mbo::enumset_exception &e) {
		printf("enumset exception\n");
	}
	printf("test finished\n");
	return 0;
}\endverbatim
	 *
	 * \return enum value
	 * \throw multi_value_exception if the set contains more than one 
	 *                              enum values
	 * \throw empty_value_exception if the set is empty
	 * \note you may use demap_value_exception to write a combined handler
	 */	
	inline TEnum to_enum() const
		throw(demap_value_exception, multi_value_exception, multi_value_exception)
	{
		return (TEnum)demap(list);
	}

	/*! Conversion to enum value
	 *
	 * When working with enum's this method returns a single enum value
	 * of the type used to instanciate this template class.
	 *
	 * \see to_enum()
	 * \return enum value
	 * \throw multi_value_exception if the set contains more than one 
	 *                              enum values
	 * \throw empty_value_exception if the set is empty
	 * \note you may use demap_value_exception to write a combined handler
	 */
	inline operator TEnum() const
		throw(demap_value_exception, multi_value_exception, multi_value_exception)
	{
		return (TEnum)demap(list);
	}
	
private:

	/*! NOT POSSIBLE - READ ABOVE
	 * \internal
	 * \note       The operator ~ is missing because the constructor cannot 
	 *             be informed about all set members available.
	 *
	 * \throw operator_na_exception allways thrown
	 */
	inline enumset & operator ~(void)
		throw(operator_na_exception)
	{
		throw operator_na_exception();
	}

	/*! Internal set representation (used as bit field)
	 */
	Ty_list list;

	/*! converts a value into mapping represenation
	 * \param value to convert
	 * \return map representation used internally
	 * \throw input_range_exception if input exceeds allowed range
	 */
	inline Ty_list map(Ty_list value) const 
		throw(input_range_exception)
	{
		if (((size_t)value) >= (sizeof(Ty_list)*8)) {
			throw input_range_exception();
		}
	    return (1L << value);
	}

	/*! converts a value into mapping represenation
	 * \param value to convert
	 * \return map representation used internally
	 * \throw multi_value_exception if the set contains more than one 
	 *                              enum values
	 * \throw empty_value_exception if the set is empty
	 * \note you may use demap_value_exception to write a combined handler
	 */
	inline Ty_list demap(Ty_list value) const 
		throw(demap_value_exception, multi_value_exception, empty_value_exception)
	{
	    register Ty_list i = 0;
	    register Ty_list p = 1;

        while (p) {
            if (list & p) {
            	if (list & (~p)) {
					throw multi_value_exception();
            	}
            	return i;
            }
            i++;
            p <<= 1;
        }
		throw empty_value_exception();
		return 0;
	}

}; // template class enumset

/****************************************************************************/
/*! \page enumset_example_1 enumset example 1
 *
 * This example shows several enumset operators in action.
 *
 * It uses the macro Execute to simplyfy the code. The macro serves two
 * two purposes. First it executes the given expression and second it prints
 * both the expression and the result using set notation through the function
 * result().
 * 
\verbatim
#include <stdio.h>
#include "enumset.hpp"

enum TEnum {a=01, b=04, c=010};
  
typedef mbo::enumset<TEnum> TSet;

char * result(const TSet & es) {
	if (!es.count())       return "[]";
	if (es==TSet(1,a))     return "[a]";
	if (es==TSet(1,b))     return "[b]";
	if (es==TSet(1,c))     return "[c]";
	if (es==TSet(2,a,b))   return "[ab]";
	if (es==TSet(2,a,c))   return "[ac]";
	if (es==TSet(2,b,c))   return "[bc]";
	if (es==TSet(3,a,b,c)) return "[abc]";
	return "illegal: more than a+b+c";
}

#define Execute(st) \
  st;\
  printf( "%-17s=> %s\n", # st, result(es))

int main(int argc, void ** argv)
{
	try {
		TSet es;                     // []
	 
		Execute( ;                ); // []
		Execute( es += a          ); // [a]    addition, union
		Execute( es  = es + b     ); // [a,b]  ditto
		Execute( es += es         ); // [a,b]  ditto
		Execute( es *= b          ); // [b]    intersection
		Execute( es  = es + c     ); // [b,c]  addition
		Execute( es -= b          ); // [c]    minus
		Execute( es.clear()       ); // []     clear
		Execute( es = TSet(0)     ); // []     dynamic creation
		Execute( es = TSet(2,a,c) ); // [a,c]  dynamic creation
		Execute( es = TEnum(es)   ); // throws multi_value_exception
	}
	catch (mbo::demap_value_exception &e) {
		printf("illegal TSet value, count != 1\n");
	}
	catch (mbo::enumset_exception &e) {
		printf("enumset exception\n");
	}
}
\endverbatim
*
* You can find another eample here \ref enumset_example_2
*
*//**************************************************************************/

/****************************************************************************/
/*! \page enumset_example_2 enumset example 2
 *
 * This example shows several enumset operators in action.
 *
 * While sd has no specific value the other sd\<n\> values have \<n\> as their
 * value after each block. This way you can simply control the results.
 *
 * The comparison operators are called from parameters of a printf call. Each 
 * of these lines show the statement to be executed, the expected result and 
 * the actual result of the operator. So again you can easily check the 
 * results.
 * 
\verbatim
#include <typeinfo>
#include <stdio.h>
#include "enumset.hpp"
       
typedef enum _Tenum {a=0, b, c} Tenum;

typedef mbo::enumset<int> Tset;

int main(int argc,char **argv)
{
	// create a set of doubles
	try {
		Tset	sd, sd0, sd1, sd2, sd3, sd4, sd5, sd6, sd7;
	
	    printf("sd0:%d\n", (size_t)sd0);
	
		sd  = Tset(a);
		sd += b;
		sd += c;
		sd -= b;
		sd5 = sd;
	
		sd  = sd + b;
		sd  = sd - c;
		sd3 = sd;
	    sd7 = sd3 + sd5;
	
	    printf("sd3:%d\n", (size_t)sd3);
	    printf("sd5:%d\n", (size_t)sd5);
	    printf("sd7:%d\n", (size_t)sd7);
	    
		printf("\n");
		sd2 = b;	
	    sd4 = c;
	    sd6 = sd2 + c;
	    sd.clear();
	    sd += a;
	    sd1 = sd0 + sd;
	    printf("sd :%d\n", (size_t)sd);
	    printf("sd1:%d\n", (size_t)sd1);
	    printf("sd2:%d\n", (size_t)sd2);
	    printf("sd4:%d\n", (size_t)sd4);
	    printf("sd6:%d\n", (size_t)sd6);

		printf("\n");
		sd2 = sd3 * sd6;
		sd4 = sd6;
		sd4*= sd5;
		sd  = sd4 * sd6;
	    printf("sd :%d\n", (size_t)sd);
	    printf("sd1:%d\n", (size_t)sd1);
	    printf("sd2:%d\n", (size_t)sd2);
	    printf("sd3:%d\n", (size_t)sd3);
	    printf("sd4:%d\n", (size_t)sd4);
	    printf("sd5:%d\n", (size_t)sd5);
	    printf("sd6:%d\n", (size_t)sd6);
	
		printf("\n");	
		printf("sd2 & sd4        == false: %s\n", sd2&sd            ? "true" : "false");
		printf("sd2 & sd3        == true : %s\n", sd2&sd3           ? "true" : "false");
		printf("sd2 & sd5        == false: %s\n", sd2&sd5           ? "true" : "false");
		printf("sd2 & a          == false: %s\n", sd2&a             ? "true" : "false");
		printf("sd2 & b          == true : %s\n", sd2&b             ? "true" : "false");
		printf("sd2 < sd3        == true : %s\n", sd2<sd3           ? "true" : "false");
		printf("sd2 < sd4        == false: %s\n", sd2<sd4           ? "true" : "false");
		printf("sd3 <= sd2       == false: %s\n", sd3<=sd2          ? "true" : "false");
		printf("sd2 >= sd2       == true : %s\n", sd2>=sd2          ? "true" : "false");
		printf("sd3 >= sd2       == true : %s\n", sd3>=sd2          ? "true" : "false");
		printf("sd2 <= sd2       == true : %s\n", sd2<=sd2          ? "true" : "false");
		printf("sd2 >= sd5       == false: %s\n", sd2>=sd5          ? "true" : "false");
		printf("sd3 >= sd2       == true : %s\n", sd3>=sd2          ? "true" : "false");
		printf("sd3 <= sd5       == false: %s\n", sd3<=sd5          ? "true" : "false");
		printf("sd5 >= sd3       == false: %s\n", sd5>=sd3          ? "true" : "false");
		printf("sd2 >= a         == false: %s\n", sd2>=a            ? "true" : "false");
		printf("sd2 >= b         == true : %s\n", sd2>=b            ? "true" : "false");
		printf("sd2 >= c         == false: %s\n", sd2>=c            ? "true" : "false");
		printf("sd1.contains(a)  == true : %s\n", sd1.contains(a)   ? "true" : "false");
		printf("sd1.contains(b)  == false: %s\n", sd1.contains(b)   ? "true" : "false");
		printf("sd1.contains(c)  == false: %s\n", sd1.contains(c)   ? "true" : "false");
		printf("sd3.contains(a)  == true : %s\n", sd3.contains(a)   ? "true" : "false");
		printf("sd3.contains(b)  == true : %s\n", sd3.contains(b)   ? "true" : "false");
		printf("sd3.contains(c)  == false: %s\n", sd3.contains(c)   ? "true" : "false");
		printf("sd4.contains(a)  == false: %s\n", sd4.contains(a)   ? "true" : "false");
		printf("sd4.contains(b)  == false: %s\n", sd4.contains(b)   ? "true" : "false");
		printf("sd4.contains(c)  == true : %s\n", sd4.contains(c)   ? "true" : "false");
		printf("sd6.contains(a)  == false: %s\n", sd6.contains(a)   ? "true" : "false");
		printf("sd6.contains(b)  == true : %s\n", sd6.contains(b)   ? "true" : "false");
		printf("sd6.contains(c)  == true : %s\n", sd6.contains(c)   ? "true" : "false");
		printf("sd3 != sd        == true : %s\n", sd3 != sd         ? "true" : "false");
		printf("sd3.contains(sd1)== true : %s\n", sd3.contains(sd1) ? "true" : "false");
		printf("sd6.contains(sd2)== true : %s\n", sd6.contains(sd2) ? "true" : "false");
		printf("sd1.contains(sd3)== false: %s\n", sd1.contains(sd3) ? "true" : "false");
		printf("sd3.contains(sd3)== true : %s\n", sd3.contains(sd3) ? "true" : "false");

	}
	catch (mbo::enumset_exception &e) {
		printf("1) %s", typeid(e).name());
	}
	
	return 0;
}
\endverbatim
*
* You can find another eample here \ref enumset_example_1
*
*//**************************************************************************/

}; // namespace mbo

#endif // _ENUMSET_HPP_
